﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace IMES_Middleware_Platform.Api.Core.Utility
{
    public static class SM3
    {
        private static readonly uint[] IV = {
        0x7380166F, 0x4914B2B9, 0x172442D7, 0xDA8A0600,
        0xA96F30BC, 0x163138AA, 0xE38DEE4D, 0xB0FB0E4E
    };

        private static uint[] T = new uint[64];
        private static uint[] W = new uint[68];
        private static uint[] W_ = new uint[64];

        static SM3()
        {
            for (int i = 0; i < 16; i++)
            {
                T[i] = 0x79CC4519;
            }
            for (int i = 16; i < 64; i++)
            {
                T[i] = 0x7A879D8A;
            }
        }

        public static byte[] ComputeHash(byte[] data)
        {
            uint[] v = (uint[])IV.Clone();
            byte[] padded = Pad(data);
            int blocks = padded.Length / 64;

            for (int i = 0; i < blocks; i++)
            {
                byte[] block = new byte[64];
                Array.Copy(padded, i * 64, block, 0, 64);
                ProcessBlock(block, ref v);
            }

            return UInt32ArrayToByteArray(v);
        }

        private static byte[] Pad(byte[] data)
        {
            int len = data.Length;
            int k = (56 - (len + 1) % 64) % 64;
            byte[] padded = new byte[len + 1 + k + 8];
            Array.Copy(data, padded, len);
            padded[len] = 0x80;
            ulong bitLength = (ulong)len * 8;
            byte[] lenBytes = BitConverter.GetBytes(bitLength);
            if (BitConverter.IsLittleEndian)
                Array.Reverse(lenBytes);
            Array.Copy(lenBytes, 0, padded, padded.Length - 8, 8);
            return padded;
        }

        private static void ProcessBlock(byte[] block, ref uint[] v)
        {
            uint[] w = new uint[68];
            uint[] w_ = new uint[64];
            uint a = v[0], b = v[1], c = v[2], d = v[3];
            uint e = v[4], f = v[5], g = v[6], h = v[7];

            for (int i = 0; i < 16; i++)
            {
                w[i] = BytesToUInt32(block, i * 4);
            }
            for (int i = 16; i < 68; i++)
            {
                w[i] = P1(w[i - 16] ^ w[i - 9] ^ RotateLeft(w[i - 3], 15))
                     ^ RotateLeft(w[i - 13], 7) ^ w[i - 6];
            }
            for (int i = 0; i < 64; i++)
            {
                w_[i] = w[i] ^ w[i + 4];
            }

            for (int i = 0; i < 64; i++)
            {
                uint ss1 = RotateLeft(RotateLeft(a, 12) + e + RotateLeft(T[i], i), 7);
                uint ss2 = ss1 ^ RotateLeft(a, 12);
                uint tt1 = FF(a, b, c, i) + d + ss2 + w_[i];
                uint tt2 = GG(e, f, g, i) + h + ss1 + w[i];
                d = c;
                c = RotateLeft(b, 9);
                b = a;
                a = tt1;
                h = g;
                g = RotateLeft(f, 19);
                f = e;
                e = P0(tt2);
            }

            v[0] ^= a;
            v[1] ^= b;
            v[2] ^= c;
            v[3] ^= d;
            v[4] ^= e;
            v[5] ^= f;
            v[6] ^= g;
            v[7] ^= h;
        }

        private static uint FF(uint x, uint y, uint z, int j)
        {
            return j < 16 ? x ^ y ^ z : (x & y) | (x & z) | (y & z);
        }

        private static uint GG(uint x, uint y, uint z, int j)
        {
            return j < 16 ? x ^ y ^ z : (x & y) | (~x & z);
        }

        private static uint P0(uint x) => x ^ RotateLeft(x, 9) ^ RotateLeft(x, 17);
        private static uint P1(uint x) => x ^ RotateLeft(x, 15) ^ RotateLeft(x, 23);

        private static uint RotateLeft(uint x, int n) => (x << n) | (x >> (32 - n));

        private static uint BytesToUInt32(byte[] bytes, int offset)
        {
            return (uint)(bytes[offset] << 24 | bytes[offset + 1] << 16
                         | bytes[offset + 2] << 8 | bytes[offset + 3]);
        }

        private static byte[] UInt32ArrayToByteArray(uint[] array)
        {
            byte[] result = new byte[array.Length * 4];
            for (int i = 0; i < array.Length; i++)
            {
                byte[] bytes = BitConverter.GetBytes(array[i]);
                if (BitConverter.IsLittleEndian)
                    Array.Reverse(bytes);
                Array.Copy(bytes, 0, result, i * 4, 4);
            }
            return result;
        }
    }

    public static class HMAC_SM3
    {
        private const int BlockSize = 64; // SM3的块大小是64字节

        public static byte[] ComputeHash(byte[] key, byte[] message)
        {
            // 处理密钥
            if (key.Length > BlockSize)
            {
                key = SM3.ComputeHash(key);
            }
            if (key.Length < BlockSize)
            {
                byte[] paddedKey = new byte[BlockSize];
                Array.Copy(key, paddedKey, key.Length);
                key = paddedKey;
            }

            byte[] ipad = new byte[BlockSize];
            byte[] opad = new byte[BlockSize];
            for (int i = 0; i < BlockSize; i++)
            {
                ipad[i] = (byte)(key[i] ^ 0x36);
                opad[i] = (byte)(key[i] ^ 0x5C);
            }

            // 计算内部哈希
            byte[] innerData = new byte[ipad.Length + message.Length];
            Buffer.BlockCopy(ipad, 0, innerData, 0, ipad.Length);
            Buffer.BlockCopy(message, 0, innerData, ipad.Length, message.Length);
            byte[] innerHash = SM3.ComputeHash(innerData);

            // 计算外部哈希
            byte[] outerData = new byte[opad.Length + innerHash.Length];
            Buffer.BlockCopy(opad, 0, outerData, 0, opad.Length);
            Buffer.BlockCopy(innerHash, 0, outerData, opad.Length, innerHash.Length);
            return SM3.ComputeHash(outerData);
        }

        public static byte[] ComputeHash(string key, string message, System.Text.Encoding encoding)
        {
            return ComputeHash(
                encoding.GetBytes(key),
                encoding.GetBytes(message)
            );
        }

    }
}
